"use strict";
var __extends = (this && this.__extends) || (function () {
    /* istanbul ignore next */ 
var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) /* istanbul ignore next */  if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        /* istanbul ignore next */ 
d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var tokens_public_1 = require("../../scan/tokens_public");
var gast_public_1 = require("../grammar/gast_public");
var utils_1 = require("../../utils/utils");
var lang_extensions_1 = require("../../lang/lang_extensions");
var keys_1 = require("../grammar/keys");
var GAstVisitor = gast_public_1.gast.GAstVisitor;
function addTerminalToCst(node, token, tokenTypeName) {
    node.children[tokenTypeName].push(token);
}
exports.addTerminalToCst = addTerminalToCst;
function addNoneTerminalToCst(node, ruleName, ruleResult) {
    node.children[ruleName].push(ruleResult);
}
exports.addNoneTerminalToCst = addNoneTerminalToCst;
var NamedDSLMethodsCollectorVisitor = (function (_super) {
    __extends(NamedDSLMethodsCollectorVisitor, _super);
    function NamedDSLMethodsCollectorVisitor(ruleIdx) {
        var _this = _super.call(this) || this;
        _this.result = [];
        _this.ruleIdx = ruleIdx;
        return _this;
    }
    NamedDSLMethodsCollectorVisitor.prototype.collectNamedDSLMethod = function (node, newNodeConstructor, methodIdx) {
        if (!utils_1.isUndefined(node.name)) {
            // copy without name so this will indeed be processed later.
            var nameLessNode = void 0;
            if (utils_1.has(node, "separator")) {
                // hack to avoid code duplication and refactoring the Gast type declaration / constructors arguments order.
                nameLessNode = new newNodeConstructor(node.definition, node.separator, node.occurrenceInParent);
            }
            else {
                nameLessNode = new newNodeConstructor(node.definition, node.occurrenceInParent);
            }
            var def = [nameLessNode];
            var key = keys_1.getKeyForAutomaticLookahead(this.ruleIdx, methodIdx, node.occurrenceInParent);
            this.result.push({ def: def, key: key, name: node.name });
        }
    };
    NamedDSLMethodsCollectorVisitor.prototype.visitOption = function (node) {
        this.collectNamedDSLMethod(node, gast_public_1.gast.Option, keys_1.OPTION_IDX);
    };
    NamedDSLMethodsCollectorVisitor.prototype.visitRepetition = function (node) {
        this.collectNamedDSLMethod(node, gast_public_1.gast.Repetition, keys_1.MANY_IDX);
    };
    NamedDSLMethodsCollectorVisitor.prototype.visitRepetitionMandatory = function (node) {
        this.collectNamedDSLMethod(node, gast_public_1.gast.RepetitionMandatory, keys_1.AT_LEAST_ONE_IDX);
    };
    NamedDSLMethodsCollectorVisitor.prototype.visitRepetitionMandatoryWithSeparator = function (node) {
        this.collectNamedDSLMethod(node, gast_public_1.gast.RepetitionMandatoryWithSeparator, keys_1.AT_LEAST_ONE_SEP_IDX);
    };
    NamedDSLMethodsCollectorVisitor.prototype.visitRepetitionWithSeparator = function (node) {
        this.collectNamedDSLMethod(node, gast_public_1.gast.RepetitionWithSeparator, keys_1.MANY_SEP_IDX);
    };
    NamedDSLMethodsCollectorVisitor.prototype.visitAlternation = function (node) {
        var _this = this;
        this.collectNamedDSLMethod(node, gast_public_1.gast.Alternation, keys_1.OR_IDX);
        var hasMoreThanOneAlternative = node.definition.length > 1;
        utils_1.forEach(node.definition, function (currFlatAlt, altIdx) {
            if (!utils_1.isUndefined(currFlatAlt.name)) {
                var def = currFlatAlt.definition;
                if (hasMoreThanOneAlternative) {
                    def = [new gast_public_1.gast.Option(currFlatAlt.definition)];
                }
                else {
                    def = currFlatAlt.definition;
                }
                var key = keys_1.getKeyForAltIndex(_this.ruleIdx, keys_1.OR_IDX, node.occurrenceInParent, altIdx);
                _this.result.push({ def: def, key: key, name: currFlatAlt.name });
            }
        });
    };
    return NamedDSLMethodsCollectorVisitor;
}(GAstVisitor));
exports.NamedDSLMethodsCollectorVisitor = NamedDSLMethodsCollectorVisitor;
// TODO: maybe just need a list of names if we use pure arrays?
function buildChildrenDictionaryDefTopRules(topRules, fullToShortName) {
    var result = new lang_extensions_1.HashTable();
    utils_1.forEach(topRules, function (currTopRule) {
        var currChildrenNames = buildChildDictionaryDef(currTopRule.definition);
        var currTopRuleShortName = fullToShortName.get(currTopRule.name);
        result.put(currTopRuleShortName, currChildrenNames);
        var namedCollectorVisitor = new NamedDSLMethodsCollectorVisitor(currTopRuleShortName);
        currTopRule.accept(namedCollectorVisitor);
        utils_1.forEach(namedCollectorVisitor.result, function (_a) {
            var def = _a.def, key = _a.key;
            var currNestedChildrenNames = buildChildDictionaryDef(def);
            result.put(key, currNestedChildrenNames);
        });
    });
    return result;
}
exports.buildChildrenDictionaryDefTopRules = buildChildrenDictionaryDefTopRules;
function buildChildDictionaryDef(initialDef) {
    var result = [];
    var possiblePaths = [];
    possiblePaths.push({ def: initialDef });
    var currDef;
    var currInIteration;
    var currInOption;
    var currResult;
    function addSingleItemToResult(itemName) {
        result.push(itemName);
        var nextPath = {
            def: utils_1.drop(currDef),
            inIteration: currInIteration,
            inOption: currInOption,
            currResult: utils_1.cloneObj(currResult)
        };
        possiblePaths.push(nextPath);
    }
    while (!utils_1.isEmpty(possiblePaths)) {
        var currPath = possiblePaths.pop();
        currDef = currPath.def;
        currInIteration = currPath.inIteration;
        currInOption = currPath.inOption;
        currResult = currPath.currResult;
        // For Example: an empty path could exist in a valid grammar in the case of an EMPTY_ALT
        if (utils_1.isEmpty(currDef)) {
            continue;
        }
        var prod = currDef[0];
        if (prod instanceof gast_public_1.gast.Terminal) {
            var terminalName = tokens_public_1.tokenName(prod.terminalType);
            addSingleItemToResult(terminalName);
        }
        else if (prod instanceof gast_public_1.gast.NonTerminal) {
            var nonTerminalName = prod.nonTerminalName;
            addSingleItemToResult(nonTerminalName);
        }
        else if (prod instanceof gast_public_1.gast.Option) {
            if (!utils_1.isUndefined(prod.name)) {
                addSingleItemToResult(prod.name);
            }
            else {
                var nextPathWith = {
                    def: prod.definition.concat(utils_1.drop(currDef)),
                };
                possiblePaths.push(nextPathWith);
            }
        }
        else if (prod instanceof gast_public_1.gast.RepetitionMandatory || prod instanceof gast_public_1.gast.Repetition) {
            if (!utils_1.isUndefined(prod.name)) {
                addSingleItemToResult(prod.name);
            }
            else {
                var nextDef = prod.definition.concat(utils_1.drop(currDef));
                var nextPath = {
                    def: nextDef
                };
                possiblePaths.push(nextPath);
            }
        }
        else if (prod instanceof gast_public_1.gast.RepetitionMandatoryWithSeparator || prod instanceof gast_public_1.gast.RepetitionWithSeparator) {
            if (!utils_1.isUndefined(prod.name)) {
                addSingleItemToResult(prod.name);
            }
            else {
                var separatorGast = new gast_public_1.gast.Terminal(prod.separator);
                var secondIteration = new gast_public_1.gast.Repetition([separatorGast].concat(prod.definition), prod.occurrenceInParent);
                // Hack: X (, X)* --> (, X) because it is identical in terms of identifying "isCollection?"
                var nextDef = [secondIteration].concat(utils_1.drop(currDef));
                var nextPath = {
                    def: nextDef
                };
                possiblePaths.push(nextPath);
            }
        }/* istanbul ignore else */ 
        else if (prod instanceof gast_public_1.gast.Alternation) {
            
            if (!utils_1.isUndefined(prod.name)) {
                addSingleItemToResult(prod.name);
            }
            else {
                // the order of alternatives is meaningful, FILO (Last path will be traversed first).
                for (var i = prod.definition.length - 1; i >= 0; i--) {
                    var currAlt = prod.definition[i];
                    // named alternatives
                    if (!utils_1.isUndefined(currAlt.name)) {
                        addSingleItemToResult(currAlt.name);
                    }
                    else {
                        var newDef = currAlt.definition.concat(utils_1.drop(currDef));
                        var currAltPath = {
                            def: newDef
                        };
                        possiblePaths.push(currAltPath);
                    }
                }
            }
        }
        else {
            /* istanbul ignore next */ throw Error("non exhaustive match");
        }
    }
    return result;
}
exports.buildChildDictionaryDef = buildChildDictionaryDef;
function initChildrenDictionary(collections) {
    var childrenDictionary = {};
    var collectionsLength = collections.length;
    for (var i = 0; i < collectionsLength; i++) {
        var key = collections[i];
        childrenDictionary[key] = [];
    }
    return childrenDictionary;
}
exports.initChildrenDictionary = initChildrenDictionary;
//# sourceMappingURL=cst.js.map